""" Simple metrics included for user convenience. For more metrics
import from scikit-learn or define your own. """

import numpy as np
from math import sqrt
from scipy.stats.stats import pearsonr


def RMSE(targets, predictions):
    """ root mean square error """
    return sqrt(((targets - predictions) ** 2).mean())

def Corr(targets, predictions):
    """ Pearson correlation coefficient """
    corr_c = pearsonr(targets, predictions)
    return corr_c[0]

def MAE(targets, predictions):
    """ mean absolute error """
    return np.absolute(targets - predictions).mean()


def MAPE(targets, predictions):
    """ mean absolute percent error """
    return np.mean(np.abs((targets - predictions) / targets))


import pandas as pd
import numpy as np
from functools import reduce

TS_VAR = 'Timestamp'


class TSConfig(object):
    """ Register and preprocess time series from arbitrary datasets.

    This simplifies the process of combining datasets from different domains.
    The data is unified into a single configuration object which is then
    handled directly by time series models. The main features include:

        1. Register each dataset so that TSConfig can automatically merge
            on timestamp. Arbitrary combinations of datasets/variables can
            be specified for modeling.
        2. Conveniently add autoregressive (lag) terms for any variable from
            any dataset, with extension to the future as far as the data goes.
            Unlike the default pandas lag function which truncates the data at
            the last date, this allows prediction on the future
            (i.e. forecasting).
        3. Simulate information lag at the variable level.
            In other words, rows can be shifted so that the prediction for each
            timestamp uses only what information would have been available at
            the forecast time.

    Example:
        Suppose we have a target variable in the dataframe 'cdc', and a
        predictor dataset in the dataframe 'external'. First register the
        data:
        >>> dc = TSConfig()
        >>> dc.register_dataset(cdc, 'CDC', 'target')
        >>> dc.register_dataset(external, 'pred', 'predictor')

        Add lag terms of the target variable as autoregressive predictors:
        >>> dc.add_AR(range(1, 7), dataset='CDC', var_names='%ILI')

        Call the stack method to combine the datasets
        >>> dc.stack()

        Combined dataframes (as an (X, y) tuple) can then be accessed using:
        >>> dc.data
    """
    def __init__(self):
        self.datasets = {}
        self.target = None
        self.predictors = []
        self.prepared_data = None
        self.forecast_delay = 0
        self.ar_set = False

    def register_dataset(self, data, name, type, var_names=None, copy=True):
        """ Register a dataset with the TSConfig.

        Inputs:
            data (pd.DataFrame): Must contain at least 2 columns, a timestamp
                column (called 'Timestamp'), and a variable column.

            name (str): Name to associate with the dataset

            type (str): 'target' or 'predictor'

            var_names (list): Optional, list of columns to keep

            copy (bool): Make a copy of data before modifying. If set to false,
                the original dataframe will be altered (only recommend
                when memory is a constraint)
        """
        if copy:
            data = data.copy()
        if type == 'target':
            data = self._register_target(data, name, var_names)
        elif type == 'predictor':
            data = self._register_predictor(data, name, var_names)
        else:
            raise ValueError("type must be 'target' or 'predictor'")
        self.datasets[name] = data

    def _register_target(self, data, name, var_names=None):
        assert TS_VAR in data.columns, 'dataset missing date var'
        data[TS_VAR] = pd.to_datetime(data[TS_VAR])
        data.sort_values(by=TS_VAR, inplace=True)
        data.set_index(TS_VAR, inplace=True)

        if var_names:
            assert isinstance(var_names, str), \
                'only a single target per dataset currently supported'
        else:
            assert data.shape[1] == 1, 'var_names must be specified for >1 col'
            var_names = data.columns[data.columns != TS_VAR][0]

        # select column as dataframe
        data = data.loc[:, [var_names]]
        self.target = name
        return data

    def _register_predictor(self, data, name, var_names=None):
        assert TS_VAR in data.columns, 'dataset missing date var'
        data[TS_VAR] = pd.to_datetime(data[TS_VAR])
        data.sort_values(by=TS_VAR, inplace=True)
        data.set_index(TS_VAR, inplace=True)

        if not var_names:
            var_names = data.columns[data.columns != TS_VAR].tolist()

        # ensure var_names is a list for dataframe selection
        if isinstance(var_names, str):
            var_names = [var_names]

        data = data.loc[:, var_names]
        self.predictors.append(name)
        return data

    def _extend_date_range(self, df, n_periods):
        """ the standard pandas shift cuts off lags at the last timestamp, but
        we instead want to extend the time index to fully fit the lags. """
        # infer time between entries
        tdelta = df.index.inferred_freq
        assert tdelta, 'Dataset has gaps or otherwise unable to infer freq'

        # add additional empty rows for extended data
        extend_range = pd.date_range(df.index[-1],
                                     periods=n_periods + 1,
                                     freq=tdelta)[1:]

        extend_df = pd.DataFrame(index=extend_range)
        new_df = pd.concat([df, extend_df], sort=True)
        return new_df

    def set_delay(self, periods, datasets='all'):
        """ Specify information delays by dataset.

        This simulates delays in receiving a data source. Here, one can
        specify which datasets are delayed. Caution: Make sure whether AR lags
        based on the target variable should be delayed based on the situation.
        Setting datasets='all' will delay these lags as well, which may not
        be realistic for knowledge delays. This setting is similar to
        forecasting with an important difference: The training set will run up
        to the present, whereas in forecasting the training set is limited to
        the most recent available target value.

        Because of the interaction with the delays and AR terms, set_delay
        methods must be called before add_AR.

        Inputs:
            periods (int): number of time intervals of delay

            datasets (str or list): list of datasets to apply delay or 'all'
                to delay all including AR lags of the target variable.
        """
        periods = int(periods)

        if self.ar_set:
            raise RuntimeError('set_delay must be used before add_AR')
        if datasets == 'all':
            datasets = self.predictors
            self.forecast_delay = periods
        else:
            assert isinstance(datasets, list), 'Pass a list of datasets.'

        for ds in datasets:
            tmp_df = self.datasets[ds].copy()
            extend_df = self._extend_date_range(tmp_df, periods)
            shifted_df = extend_df.shift(periods)

            shifted_df.index.rename(TS_VAR, inplace=True)
            self.datasets[ds] = shifted_df

    def add_AR(self, terms, dataset, var_names='all'):
        """ Creates an autoregressive (lagged) dataset as additional features.
        These are stored using the name 'AR_dataset'.

        Inputs:
            terms (list): lag terms to consider, e.g. 1 means 1 period ago.

            dataset (str): Dataset containing the variables to be lagged

            var_names (str or list): variables to lag. 'all' to select all.
        """
        self.ar_set = True

        # ensure var_names is a list for dataframe selection
        if isinstance(var_names, str):
            if var_names == 'all':
                var_names = self.datasets[dataset].columns
            else:
                raise ValueError("Pass 'all' or a list of variables.")

        # if lag is for target, add forecasting delay to lag
        # other datasets already have delay applied.
        if dataset == self.target:
            shift = max(terms) + self.forecast_delay
            terms = [x + self.forecast_delay for x in terms]
        else:
            shift = max(terms)

        base_df = self.datasets[dataset].loc[:, var_names]
        start_data = self._extend_date_range(base_df, shift)

        # concatenate all lags
        ar_predictors = pd.concat(
            [start_data.shift(x).add_suffix('_lag{0}'.format(x)) for x in terms],
            axis=1)

        ar_predictors.index.rename(TS_VAR, inplace=True)
        name = 'AR_{0}'.format(dataset)
        self.datasets[name] = ar_predictors
        self.predictors.append(name)

    def stack(self, predictors='all', merge_type='outer', fill_na='ignore'):
        """ Merge datasets together into final modeling dataframe

        Inputs:
            predictors ('all' or list): All predictor datasets to use

            merge_type (str): pandas merge option. In general use 'outer'.

            fill_na (str): How to handle missing data.
        """
        if predictors == 'all':
            dfs = [self.datasets[x] for x in self.predictors]
        elif isinstance(predictors, list):
            dfs = [self.datasets[x] for x in predictors]
        self.prepared_data = reduce(
            lambda x, y: pd.merge(x, y, left_index=True, right_index=True,
                                  how=merge_type), dfs)
        self.prepared_data.sort_index(inplace=True)
        if fill_na != 'ignore':
            raise NotImplementedError

    @property
    def data(self):
        if self.prepared_data is None:
            raise RuntimeError('Call stack method before returning data.')
        return self.prepared_data, self.datasets[self.target]


import numpy as np
import pandas as pd
from tqdm import tqdm
from nowcast import TSConfig


class Arex(object):
    """ AREX (AutoRegression with EXogeneity) is an iterative time series
    predictor that abstracts away the logic of retraining a model sequentially
    on time series data. AREX does not impose any modeling constraints --
    instead it is a procedure that can handle any model that is compatible
    with scikit-learn's fit/predict API.

    Usually, one retrains a time series model at each time step in order to use
    the most recent information. The training set at each step can be either
    rolling (fixed size that discards old data), or expanding (use all data).
    Often, a time series is predicted using a combination of lags of the
    time series (AR), concurrent exogenous variables (EX), and lags of the
    exogenous variables. AREX takes care of these details for you.

    On the other hand, the actual model that is applied at each time step is
    highly important to researchers -- it can involve preprocessing and feature
    engineering to using various ML algorithms and hyperparameter tuning. Thus
    this part is flexible and only limited by your creativity. All AREX needs
    is a model class with fit() and predict() methods, identical to sklearn.
    In fact, any sklearn model can be passed directly into AREX to get an
    out-of-the-box time series modeler.

    Example:
        Suppose we continue the example with TSConfig.
        >>> dc = TSConfig()
        >>> dc.register_dataset(cdc, 'CDC', 'target')
        >>> dc.register_dataset(external, 'pred', 'predictor')
        >>> dc.add_AR(range(1, 7), dataset='CDC', var_names='%ILI')
        >>> dc.stack()

        We will use a default sklearn random forest as the model:
        >>> mod = RandomForestRegressor()

        The above time series is at a weekly frequency. For nowcasting
        (predicting target at week t using exogenous data from week t) with
        a year-long rolling training window, do:
        >>> arex = Arex(model=mod, data_config=dc)
        >>> pred = arex.nowcast(pred_start='2019-02-19', pred_end='2019-08-20',
                                training='roll', window=52)

        Suppose we want to predict a week ahead. We would do:
        >>> pred2 = arex.forecast(t_plus=1, pred_start='2019-02-19',
                                  pred_end='2019-08-20',
                                  training='roll', window=52)

        Note that the timestamps for pred_start and pred_end refer to the time
        of making the prediction, not the time that is predicted.

    Inputs:
        model (class): Any model with fit and predict methods following sklearn
        API. The methods must accept pandas dataframes.

        X (dataframe): Predictor dataframe. Either pass X and y or pass a
        TSConfig object to data_config.

        y (dataframe): Target dataframe that should consist of a single column.

        data_config (TSConfig): Preprocessed X and y data using TSConfig. This
        will overwrite X and y if passed.

        verbose (int): Control verbosity of output
    """
    def __init__(self, model, X=None, y=None, data_config=None, verbose=1):
        self.model = model
        self.X = X
        self.y = y
        self.config = data_config
        self.pred_start = None
        self.pred_end = None
        self.training = None
        self.window = None
        self.pred_name = None
        self.verbose = verbose
        self.log = []

        self._validate_init()

        if self.config:
            self.X, self.y = self.config.data

        if self.verbose > 0:
            print('---------------- ArEx model initialized ----------------')

    def forecast(self, t_plus, pred_start, pred_end,
                training, window, pred_name='Predicted', t_known=False):
        """ Performs rolling forecast on data.

        Inputs:
            t_plus (int): Number of periods ahead to forecast. 0 corresponds to
                nowcasting (predict y_t from X_t).

            pred_start (str or datetime): Time of first prediction

            pred_end (str or datetime): Time of last prediction

            training (str): Training window behavior, either 'roll' or 'expand'

            window (int): Training window size. If training is 'expand', then
                window determines the size for the first prediction, and
                subsequent windows increase in size.

            pred_name (str): Name for prediction output column

            t_known (bool): Whether y_t would be known at time of forecasting.
                This adjusts the training window to use the most recent
                information. Since y_t is the nowcast target, t_known must be
                set to False when t_plus=0.
        """
        self.t_plus = t_plus
        self.pred_start = pd.to_datetime(pred_start)
        self.pred_end = pd.to_datetime(pred_end)
        self.training = training
        self.window = window
        self.pred_name = pred_name
        self.t_known = t_known
        self.log = []

        self._validate_predict()

        if self.verbose > 0:
            print('Predicting from {0} to {1}:'.format(self.pred_start,
                                                       self.pred_end))

        # range of timestamps to predict
        pred_range = self.X.loc[self.pred_start:self.pred_end].index

        predictions = []
        for timestamp in tqdm(pred_range):
            X_train_start, X_train_end, y_train_start, y_train_end = \
                self._get_start_end_index(timestamp)

            X_train = self.X.iloc[X_train_start:X_train_end]
            y_train = self.y.iloc[y_train_start:y_train_end, 0]

            X_pred_index = self._get_pred_index(timestamp)
            X_pred = self.X.iloc[X_pred_index:X_pred_index + 1]

            if self.verbose == 2:
                print('Predicting time: ', timestamp)
                print('X_train: ', X_train.index[0], X_train.index[-1])
                print('y_train: ', y_train.index[0], y_train.index[-1])
                print('X_pred: ', X_pred.index[0])
                print('sizes: ', X_train.shape, y_train.shape)

            debug_log = {}
            debug_log['time'] = timestamp
            debug_log['X_train'] = (X_train.index[0], X_train.index[-1])
            debug_log['y_train'] = (y_train.index[0], y_train.index[-1])
            debug_log['X_pred'] = X_pred.index[0]
            debug_log['sizes'] = (X_train.shape, y_train.shape)
            self.log.append(debug_log)

            self.model.fit(X_train, y_train)
            predictions.append(self.model.predict(X_pred)[0])

        pred_df = pd.DataFrame(data={self.pred_name: predictions},
                               index=pred_range).reset_index()
        return pred_df

    def nowcast(self, pred_start, pred_end,
                training, window, pred_name='Predicted'):
        """ Perform rolling nowcast on data. This is just a convenience wrapper
        for the forecast method.
        """
        return self.forecast(t_plus=0, pred_start=pred_start,
                             pred_end=pred_end, training=training,
                             window=window, pred_name=pred_name,
                             t_known=False)

    def get_params(self):
        """ Returns user-set parameters of Arex object """
        params = {}
        params['pred_start'] = self.pred_start
        params['pred_end'] = self.pred_end
        params['training'] = self.training
        params['window'] = self.window
        params['pred_name'] = self.pred_name
        return params

    def get_log(self):
        """ Return a dataframe of metadata for each prediction. For example,
        the X_train and y_train date ranges, prediction time, and training set
        sizes. Use this for debugging or to confirm behavior.
        """
        return pd.DataFrame(self.log)

    def _validate_init(self):
        assert hasattr(self.model, 'fit') and hasattr(self.model, 'predict'), \
            'model must have fit and predict methods.'

        if self.X is not None and self.y is not None:
            assert isinstance(self.X, pd.DataFrame), \
                'X and y must be dataframes'
            assert isinstance(self.y, pd.DataFrame), \
                'X and y must be dataframes'

        if self.config:
            assert isinstance(self.config, TSConfig), \
                'data_config must be a TSConfig object'

        assert self.config or (self.X is not None and self.y is not None), \
            'Either pass X and y dataframes or a TSConfig object'
        if self.config and (self.X and self.y):
            print('Both X and y and TSConfig were passed, will default to '
                  'TSConfig')

    def _validate_predict(self):
        assert self.training in ('expand', 'roll')
        if self.training == 'roll':
            assert self.window > 0
            self.window = int(self.window)

        assert self.pred_start in self.X.index, 'pred_start not in X dataframe'
        assert self.pred_end in self.X.index, 'pred_end not in X dataframe'

    def _get_start_idx(self, pred_timestamp):
        # forecasting shifts train set back
        backshift = self.t_plus - int(self.t_known)

        if pred_timestamp in self.y.index:
            # if target is available at train end
            X_train_end = np.where(self.X.index == pred_timestamp)[0][0]
            y_train_end = np.where(self.y.index == pred_timestamp)[0][0]
            X_train_end = X_train_end - backshift
            y_train_end = y_train_end + int(self.t_known)
        else:
            # find most recent available training timestamp
            last_train_ts = min(self.X.index.max(), self.y.index.max())
            X_train_end = np.where(self.X.index == last_train_ts)[0][0] + 1
            y_train_end = np.where(self.y.index == last_train_ts)[0][0] + 1
            X_train_end = X_train_end - self.t_plus
        return X_train_end, y_train_end

    def _get_start_end_index(self, pred_timestamp):

        # find train end parameters
        X_train_end, y_train_end = self._get_start_idx(pred_timestamp)

        # find train start parameters
        if self.training == 'expand':
            X_pred_start, y_pred_start = self._get_start_idx(self.pred_start)
            X_train_start = X_pred_start - self.window
            y_train_start = y_pred_start - self.window

        else:
            X_train_start = X_train_end - self.window
            y_train_start = y_train_end - self.window

        if min(X_train_start, y_train_start) < 0:
            raise ValueError(
                'Train set has {0} out of {1} rows needed'.format(
                    min(X_train_end, y_train_end), self.window))

        return X_train_start, X_train_end, y_train_start, y_train_end

    def _get_pred_index(self, pred_timestamp):
        return np.where(self.X.index == pred_timestamp)[0][0]



import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LassoCV
from sklearn.svm import SVR
from sklearn.model_selection import GridSearchCV
from scipy.special import logit, expit


class Argo(object):
    """ The base ARGO model without extra features, with Lasso regression """
    def __init__(self):
        self.model = LassoCV(eps=0.001, n_alphas=200, cv=10,
                             max_iter=1e4, n_jobs=-1)
        self.scaler = StandardScaler()

    def fit(self, X, y):
        X_train = self.scaler.fit_transform(X)
        self.model.fit(X_train, y)

    def predict(self, X):
        X_pred = self.scaler.transform(X)
        return self.model.predict(X_pred)


class Argo2015(object):
    """ The ARGO model based on Yang et al. 2015

        Currently not configured to transform Google Trends data.
        It takes the logit of all inputs.
    """
    def __init__(self):
        self.model = LassoCV(eps=0.001, n_alphas=200, cv=10,
                             max_iter=1e4, n_jobs=-1)
        self.scaler = StandardScaler()

        self.logit = lambda x: logit(x)
        self.expit = lambda x: expit(x)
        self.log = lambda x: np.log(x)

    def fit(self, X, y):
        # ar_cols = [x for x in X.columns if 'AR_' in x]
        # X_ar = X[ar_cols].copy()
        # X[ar_cols] = self.logit(X_ar / 100)

        Xl = logit(X / 100)
        yl = logit(y / 100)
        X_train = self.scaler.fit_transform(Xl)
        self.model.fit(X_train, yl)

    def predict(self, X):
        Xl = self.logit(X / 100)
        X_pred = self.scaler.transform(Xl)
        return self.expit(self.model.predict(X_pred)) * 100


class ArgoSVM(object):
    """ ARGO with SVM regressor, based on Santillana et al. 2016 """
    def __init__(self):
        import warnings
        warnings.filterwarnings("ignore")
        warnings.filterwarnings(action='ignore', category=DeprecationWarning)
        warnings.filterwarnings(action='ignore', category=FutureWarning)

        C_range = 10.0 ** np.arange(-3, 3)
        gamma_range = 10.0 ** np.arange(-3, 3)
        epsilon_range = 10.0 ** np.arange(-1, 1)
        self.params = dict(kernel=['rbf'],
                           epsilon=epsilon_range,
                           gamma=gamma_range,
                           C=C_range)
        self.model = GridSearchCV(SVR(), self.params, cv=10, n_jobs=-1)
        self.scaler = StandardScaler()

    def fit(self, X, y):
        Xl = logit(X / 100)
        yl = logit(y / 100)
        X_train = self.scaler.fit_transform(Xl)
        self.model.fit(X_train, yl)

    def predict(self, X):
        Xl = logit(X / 100)
        X_pred = self.scaler.transform(Xl)
        return expit(self.model.predict(X_pred)) * 100



import setuptools

VERSION = '0.1.1'
SHORT_DESCRIPTION = 'Light, modular framework for dynamic time series modeling'
URL = 'https://github.com/fl16180/nowcast'

setuptools.setup(
    name='nowcast',
    version=VERSION,
    author='Fred Lu',
    author_email='fredlu.flac@gmail.com',
    description=SHORT_DESCRIPTION,
    long_description='See GitHub for description',
    url=URL,
    license='GPL-3.0',
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy>=1.17.1',
        'scipy>=1.2.1',
        'scikit-learn>=0.20.3',
        'pandas>=0.25.1',
        'tqdm>=4.36.1'
    ]
)



import numpy as np
import pandas as pd
from nowcast import TS_VAR

CDC_REGIONS = './Regions.csv'
MMWR_LOOKUP = './MMWR_lookup.csv'


class CDCLoader(object):
    """ Loader for CDC data.

    Example usage:
        >>> cdcl = CDCLoader("./ILI_national_dated.csv")
        >>> cdc = cdcl.load_national()
    """
    DATE_VAR = 'Date'
    ILI_WEIGHTED = '% WEIGHTED ILI'
    ILI_UNWEIGHTED = '%UNWEIGHTED ILI'
    REGION_ID = 'REGION'
    STATE_ID = 'State'
    ILI_RENAME = "%ILI"

    def __init__(self, filename, ili_version='weighted'):
        self.data = pd.read_csv(filename)
        if ili_version == 'weighted':
            self.data[self.ILI_RENAME] = self.data[self.ILI_WEIGHTED]
        elif ili_version == 'unweighted':
            self.data[self.ILI_RENAME] = self.data[self.ILI_UNWEIGHTED]

    def load_national(self):
        data = self.data
        data[TS_VAR] = pd.to_datetime(data[self.DATE_VAR])
        return data[[TS_VAR, self.ILI_RENAME]].copy()

    def load_regional(self, region):
        data = self.data
        data[TS_VAR] = pd.to_datetime(data[self.DATE_VAR])
        data[self.REGION_ID] = data[self.REGION_ID].map(lambda x: x[7:])
        data = data[data[self.REGION_ID] == region]
        return data[[TS_VAR, self.ILI_RENAME]].copy()

    def load_state(self, state):
        data = self.data
        data[TS_VAR] = pd.to_datetime(data[self.DATE_VAR])
        data = data[data[self.STATE_ID] == state]
        return data[[TS_VAR, self.ILI_RENAME]].copy()


class AthenaLoader(object):
    """ Loader for CDC data.

    Example usage:
        >>> athl = AthenaLoader("./ATHdata.csv")
        >>> ath = athl.load_national()
    """
    DATE_VAR = 'Week Start Date'
    ATHENA_VARS = ['Flu Visit Count', 'ILI Visit Count',
                   'Unspecified Viral or ILI Visit Count']
    COUNTS = 'Visit Count'

    def __init__(self, filename, smoothing=None):
        self.data = pd.read_csv(filename)
        self.data[TS_VAR] = pd.to_datetime(self.data[self.DATE_VAR])
        if self.DATE_VAR != TS_VAR:
            self.data.drop(self.DATE_VAR, axis=1, inplace=True)
        self.region_data = None
        self.smoothing = smoothing

    def load_national(self):
        data = self.data[self.data['State'] == 'ALL STATES'].copy()
        return self._process_data(data)

    def load_regional(self, region):
        # merge with CDC region lookup
        if not self.region_data:
            regions = pd.read_csv(CDC_REGIONS)
            region_data = pd.merge(self.data, regions, how='left', on='State')
            region_data = region_data.groupby(['Region', 'Year', 'MMWR Week'],
                                              as_index=False).sum()
            self.region_data = region_data

        data = self.region_data[self.region_data['Region'] == region].copy()
        return self._process_data(data)

    def load_state(self, state):
        data = self.data[self.data['State'] == state].copy()
        return self._process_data(data)

    def _process_data(self, data):
        if self.smoothing == 'moving_avg':
            raise NotImplementedError
        else:
            data[self.ATHENA_VARS] = \
                data[self.ATHENA_VARS].div(data[self.COUNTS], axis=0)
        return data[[TS_VAR] + self.ATHENA_VARS].copy()


def gt_loader(filename):
    """ Loader for Google Trends data """
    data = pd.read_csv(filename)
    data[TS_VAR] = pd.to_datetime(data['date'])
    data.drop('date', axis=1, inplace=True)
    return data
